/*
**    BPALogin - lightweight portable BIDS2 login client
**    Copyright (c) 2001-3 Shane Hyde, and others.
** 
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/ 

/**
 * Changes:
 * 2001-09-19:  wdrose    Fixed incorrect use of single fork() to put
 *                        BPALogin into background.  Replaced with
 *                        fork(), setsid(), fork().
 *
 * 2001-12-05:  wdrose    Added fix gleaned from Sam Johnston to include
 *                        errno.h for errno, rather than assuming it is an
 *                        extern int.
 *
 *
 * 2008-10-08:  DNI's seaturn  Sorry for the stupid modification to support dial-on-demand
 *			   function. ^_^ 
 *			   NOTE: Please check __LITTLE_ENDIAN. I'm not good at configuring script.
 */

#include "bpalogin.h"

#define BPALOGIN_BANNER \
        "BPALogin v2.0.2 - portable BigPond Broadband login client"

static struct session conn;

static void on_signal(int i)
{
	struct session *s = &conn;

	DEBUGP("Signal caught\n");
	bpa_logout(0, s);
	set_bpa_mode(s, 1, 0);

	syslog(6, "[Internet disconnected]");

	exit(0);
}

int uptime(void)
{
	struct sysinfo info;

	sysinfo(&info);

	return info.uptime;
}

/*
  * returns true if the string passed in is an internet address in dotted quad
  * notation.
  * If it is as like 'test.com', NOT 'login-server', return true also.
  */
inline int is_dotted_ip(char *addr)
{
	return inet_addr(addr) != INADDR_NONE ||strchr(addr, '.') != NULL;
}

void show_bpa_status(int up)
{
	FILE *fp;

	fp = fopen("/tmp/bpa_info", "w");
	if (fp == NULL) {
		printf("Can not creat BigPond link status info!\n");
		return;
	}	
	
	if (up) {
		fprintf(fp, "up time: %d", uptime());
		system("/sbin/ledcontrol -n wan -c green -s on");
	} else {
		fprintf(fp, "down time: %d", uptime());
		system("/sbin/ledcontrol -n wan -c amber -s on");
	}

	fclose(fp);
}

static void usage( void )
{
	printf(BPALOGIN_BANNER "\n"
		"Copyright (c) 2001-3 Shane Hyde and others\n\n"
		"This program is *not* a product of Big Pond Advance\n\n"
		"Usage: bpalogin [-c file]\n\n"
		" -c file            Specifies the configuration file to use\n"
		"                    (default is " DEFAULT_CONFFILE ")\n\n"
		"Command line options override the values in the configuration "
		"file\n");
	exit(-1);
}

static int parse_parms(struct session *s, char *conffile)
{
	FILE *fp;
	char buf[512];

	fp = fopen(conffile, "rt");
	if (fp == NULL) {
		DEBUGP("Cannot open conf file\n");
		return 0;
	}

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char parm[256];
		char value[256];

		if (buf[0] == '#')
			continue;

		/**
		 * Fixed bug 16552.
		 * While users leave blank for bigpond authentication server field
		 * and fill in "aaa" in password field, DUT will query the domain
		 * as "aaa.bigpond.net.au". This is cause of the previous string in
		 * variable "value" was not replaced to blank.
		 * solution:
		 * 	set character array "parm" and "value" to blank before get these values.
		 */
		memset(parm, 0, sizeof(parm));
		memset(value, 0, sizeof(value));

		/*
		  * Problem with using sscanf(buf, "%s %s"), parm, value):
		  * usernames with periods et al are not picked up correctly.
		  * Really need to use strtok.
		  */
		sscanf(buf, "%s %s", parm, value);    
		DEBUGP("Parameter %s set to %s\n", parm, value);

		if (!strcasecmp(parm,"username")) {
			strcpy(s->username, value);
		}
		else if (!strcasecmp(parm,"password")) {
			strcpy(s->password, value);
		}
		else if (!strcasecmp(parm,"authdomain")) {
			strcpy(s->authdomain, ".");
			strcat(s->authdomain, value);
		}
		else if (!strcasecmp(parm,"authserver")) {
			strcpy(s->authserver, value);
		}
		else if (!strcasecmp(parm,"localaddress")) {
			strcpy(s->localaddress, value);
		}
		else if (!strcasecmp(parm, "devname")) {
			strcpy(s->devname, value);
		}
		else if (!strcasecmp(parm, "linkmode")) {
			if (!strcasecmp(value, "demand"))
				s->linkmode = DEMAND;
			else if (!strcasecmp(value, "runone"))
				s->linkmode = RUNONE;
			else if (!strcasecmp(value, "persist"))
				s->linkmode = PERSIST;
		}
		else if (!strcasecmp(parm,"minheartbeatinterval")) {
			s->minheartbeat = atoi(value);
		}
		else if (!strcasecmp(parm,"maxheartbeatinterval")) {
			s->maxheartbeat = atoi(value);
		}
		else if (!strcasecmp(parm,"idletime")) {
			s->idletime = (time_t)atoi(value);
		}
		else if (!strcasecmp(parm,"localport")) {
			s->localport = atoi(value);
		}
	}

	fclose(fp);

	/* 
	  * [NETGEAR SPEC V1.6] 5.5 PPP dial on demand:
	  * If Idle Timeout value is 0, it means the PPPoE, PPTP or BPA connection will always keep alive.
	  */
	if (s->linkmode == DEMAND && s->idletime == 0)
		s->linkmode = PERSIST;

	/*
	  * [NETGEAR SPEC V1.6] 5.2 BPA:
	  * Router SHOULD detect the domain information from DHCP server and bind
	  * the auth server prefix "login-server" as the authentication server to initial
	  * the authentication steps. Normally the auth server is:
	  * `login-server.<dhcp-learned-domain-suffix>`. Here the
	  * `<dhcp-learned-domain-suffix>` is different in different province.
	  */
	if (!is_dotted_ip(s->authserver))
		strcat(s->authserver, s->authdomain);

	return 1;
}

int main(int argc,char* argv[])
{
	int c;
	char *mode;
	struct session *s;
	char conffile[256];

	signal(SIGINT, on_signal);
	signal(SIGHUP, on_signal);
	signal(SIGTERM, on_signal);

	s = &conn;
	strcpy(s->authserver, DEFAULT_AUTHSERVER);
	strcpy(s->authdomain, DEFAULT_AUTHDOMAIN);
	strcpy(s->devname, DEFAULT_DEVNAME);
	strcpy(conffile, DEFAULT_CONFFILE);

	/* ACCEPT          net       fw            udp     520,5050 */
	s->localport = 5050;
	s->minheartbeat = 60;
	s->maxheartbeat = 420;
	s->authport = DEFAULT_AUTHPORT;
	s->linkmode = PERSIST;

	while ((c = getopt( argc, argv, "c:" )) > -1) {
		switch( c ) {
		case 'c':
			strncpy(conffile, optarg, sizeof(conffile));
			break;
		case '?':
			usage();
   			break;
		}
	}

	if (!parse_parms(s, conffile)) {
		printf("bpalogin: Could not read configuration file (%s)\n\n", conffile);
		usage();
	}

	if (s->username[0] == '\0') {
		DEBUGP("Username has not been set!!!\n");
		usage();
	}

	if (s->linkmode == RUNONE)
		mode = "Manually Connect";
	else if (s->linkmode == PERSIST)
		mode = "Always On";
	else
		mode = "Dial on Demand";
	
	daemon(1, 1);
	printf("Telstra BigPond is running on \"%s\" mode ...\n", mode);

	/**
	 * DUT should not automatically connect to Internet as BPA "dial on demand"
	 * Until valid traffic turn up.
	 */
	show_bpa_status(0);
	set_bpa_mode(s, 0, 0);
	if (s->linkmode == DEMAND)
		bpa_demand(s);

	while (mainloop(s));

	set_bpa_mode(s, 1, 0);

	exit(0);
}

