/*
**	BPALogin - lightweight portable BIDS2 login client
**	Copyright (c) 2001-3 Shane Hyde, and others.
** 
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/ 

#include "bpalogin.h"

#define GETSHORT(s, cp) { \
	(s) = *(cp)++ << 8; \
	(s) |= *(cp)++; \
}

#define GETLONG(l, cp) { \
	(l) = *(cp)++ << 8; \
	(l) |= *(cp)++; (l) <<= 8; \
	(l) |= *(cp)++; (l) <<= 8; \
	(l) |= *(cp)++; \
}

/**************************************************************************************/

static void add_trans_data(struct transaction *t, void *p, int count)
{
	memcpy(t->data + t->length, p, count);
	t->length += count;
}

static void add_int2field(struct transaction *t, uint16 v)
{
	add_trans_data(t, &v, 2);
}

static void add_int4field(struct transaction *t, uint32 v)
{
	add_trans_data(t, &v, 4);
}

static void add_string(struct transaction *t, char *str)
{
	int len;

	len = strlen(str);
	add_int2field(t, htons((uint16)(len + 4)));
	add_trans_data(t, str, len);
}

static void add_data(struct transaction *t, char *p, int c)
{
	add_int2field(t, htons((uint16)(c + 4)));
	add_trans_data(t, p, c);
}

void add_field_string(struct transaction *t, uint16 fn, char *p)
{
	add_int2field(t, htons(fn));
	add_string(t, p);
}

void add_field_data(struct transaction *t, uint16 fn, char *p, int c)
{
	add_int2field(t, htons(fn));
	add_data(t, p, c);
}

void add_field_int2(struct transaction *t, uint16 fn, uint16 v)
{
	add_int2field(t, htons(fn));
	add_int2field(t, htons(6));
	add_int2field(t, htons(v));
}

void add_field_int4(struct transaction *t, uint16 fn, uint32 v)
{
	add_int2field(t, htons(fn));
	add_int2field(t, htons(8));
	add_int4field(t, htonl(v));
}

/**************************************************************************************/

static unsigned char *locate_field(struct transaction *t, uint16 parm)
{
	int i;
	uint16 type, len;
	unsigned char *p;

	/* skip message-type(2B) + length-value(2B) + session-id(4B) */
	i = 8; p = t->data + 8;
	while (i < t->length) {
		GETSHORT(type, p);
		GETSHORT(len, p);

		if (parm == type)
			return p;

		i += len; p += (len - 4);
	}

	return NULL;
}

int extract_int2_value(struct transaction *t, uint16 parm, uint16 *v)
{
	uint16 val;
	unsigned char *p;

	p = locate_field(t, parm);
	if (p != NULL) {
		GETSHORT(val, p);
		*v = val;
		DEBUGP("Received parm %d value %d\n", parm, val);

		return 1;
	}

	return 0;
}

int extract_int4_value(struct transaction *t, uint16 parm, uint32 *v)
{
	uint32 val;
	unsigned char *p;

	p = locate_field(t, parm);
	if (p != NULL) {
		GETLONG(val, p);
		*v = val;
		DEBUGP("Received parm %d value %d\n", parm, val);

		return 1;
	}

	return 0;
}

int extract_string_value(struct transaction *t, uint16 parm, char *v)
{
	uint16 len;
	unsigned char *p;

	p = locate_field(t, parm);
	if (p != NULL) {
		p -= 2;	
		GETSHORT(len, p);
		memcpy(v, p, len - 4);
		*(v + len -4) = '\0';

		DEBUGP("Received parm %d value %s\n", parm, v);

		return 1;
	}

	return 0;
}

/**************************************************************************************/

void start_transaction(struct transaction *t, uint16 msgtype, uint32 sessionid)
{
	t->length = 0;

	memset(t->data, 0, MAX_PKTSIZE);
	add_int2field(t, htons(msgtype));
	add_int2field(t, htons(0));
	add_int4field(t, htonl(sessionid));
}

void send_transaction(int socket, struct transaction *t)
{
	uint16 *plen;

	plen = (uint16 *)(t->data + 2);
	*plen = htons((uint16)t->length);

	send(socket, t->data, t->length, 0);
}

void send_udp_transaction(struct session *s, struct transaction *t)
{
	uint16 *plen;

	plen = (uint16 *)(t->data + 2);
	*plen = htons((uint16)t->length);

	s->fromaddr.sin_port = htons(s->statusport);
	sendto(s->listensock, t->data, t->length, 0, (struct sockaddr *)&s->fromaddr,
				sizeof(s->fromaddr));
}

uint16 receive_transaction(int socket, struct transaction *t)
{
	int r;
	uint16 *v;

	r = recv(socket, t->data, MAX_PKTSIZE, 0);

	t->length = r;

	v = (uint16 *)t->data;

	return r > 0 ? ntohs(*v) : 0;
}

static int check_heartbeat_packet(struct transaction *t, int length)
{
	uint16 type;

	type = ntohs(*(uint16 *)t->data);
	if (type != T_MSG_STATUS_REQ) {
		DEBUGP("Incorrect heartbeat transaction type %d\n", type);
		return 0;
	}

	if (length != 8) {
		DEBUGP("Incorrect heartbeat transaction length %d\n", length);
		return 0;
	}

	return 1;
}

uint16 receive_udp_transaction(struct session *s, int socket, struct transaction *t, struct sockaddr_in *addr)
{
	int r, i;
	uint16 *v;
	int timo_sec;
	fd_set readfds;
	socklen_t addrlen;
	struct timeval timo;

	timo_sec = s->lastheartbeat + s->maxheartbeat - uptime();
	if (timo_sec < 1) {
		DEBUGP("BPA hasn't detected heartbeats for a period of %d minutes!\n",
			s->maxheartbeat);
		return T_STATUS_TIMEOUT;
	}

	if (s->linkmode == DEMAND) {
		unsigned long idle = bpa_idle_time(s);
		DEBUGP("BPA Idle time is %lu!\n", idle);

		if (idle != 0xFFFFFFFF) {
			int left = s->idletime - (int)idle;
			if (left < 1) {
				DEBUGP("BPA has reached the Idle time!!!\n");

				syslog(6, "[Internet idle-timeout]");
				return T_STATUS_TIMEOUT;
			}

			if (left < timo_sec)
				timo_sec = left;
		}
	}

	addrlen = sizeof(struct sockaddr_in);

	timo.tv_sec = timo_sec;
	timo.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(socket, &readfds);

	r = select(socket + 1, &readfds, NULL, NULL, &timo);
	if (r == -1)
		return T_STATUS_ERROR;

	/* Time-Out */
	if (r == 0) {
		if (s->linkmode == DEMAND)
			return T_STATUS_IGNORE_ERR; /* continue ... */
		else
			return T_STATUS_TIMEOUT;
	}
	r = recvfrom(socket, t->data, MAX_PKTSIZE, 0, (struct sockaddr*)addr, &addrlen);
	if (r <= 0)
		return T_STATUS_TIMEOUT;

	if (s->lastheartbeat + s->minheartbeat > uptime()) {
		if (++s->recenthb > 3) {
			DEBUGP("Heartbeats arriving too quickly - discarding\n");
			return T_STATUS_IGNORE_ERR;
		}
	} else {
		s->recenthb = 0;
	}

	s->lastheartbeat = uptime();

	for (i = 0; i < s->tsmcount; i++) {
		if (addr->sin_addr.s_addr == s->tsmlist_in[i].sin_addr.s_addr)
			break;
	}

	if (i == s->tsmcount) {
		DEBUGP("Received a heartbeat from unexpected source %s:%d\n",
					inet_ntoa(addr->sin_addr), ntohs(addr->sin_port));
		return T_STATUS_IGNORE_ERR;
	}

	t->length = r;

	/* Lets make sure this packet is structured correctly. */
	if (!check_heartbeat_packet(t, r))
		return T_STATUS_IGNORE_ERR;

	v = (uint16 *)t->data;

	return ntohs(*v);
}

