/*
<:copyright-gpl
 Copyright 2002 Broadcom Corp. All Rights Reserved.

 This program is free software; you can distribute it and/or modify it
 under the terms of the GNU General Public License (Version 2) as
 published by the Free Software Foundation.

 This program is distributed in the hope it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.

 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
:>
*/
/******************************************************************************
//
//  Filename:       ip_conntrack_proto_esp.c
//  Author:         Pavan Kumar
//  Creation Date:  05/27/04
//  Modified by Vincent Yang at DNI. Taiwan 2009/11/27
//
//  Description:
//      Implements the ESP ALG connectiontracking.
//
*****************************************************************************/
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>

#include "ip_conntrack_esp.h"

static struct _esp_table esp_table[MAX_PORTS];
static u_int16_t cur_spi = 0;

static void
esp_free_entry(int index)
{
    if (esp_table[index].inuse) {
        if (esp_table[index].timer_active) {
            DEBUGP("Deleting timer index %d\n", index);
            del_timer(&esp_table[index].refresh_timer);
        }
        memset(&esp_table[index], 0, sizeof(struct _esp_table));
    }
}

static void
esp_refresh_ct(unsigned long data)
{
    struct _esp_table *esp_entry = NULL;

    if (data > MAX_PORTS) {
        return;
    }

    esp_entry = &esp_table[data];
    if ( esp_entry == NULL ) {
        return;
    }
    DEBUGP("ntimeouts %d pkt_rcvd %d entry %p data %lu ct %p\n",
           esp_entry->ntimeouts, esp_entry->pkt_rcvd, esp_entry, data, esp_entry->ct);
    if (esp_entry->pkt_rcvd) {
        esp_entry->pkt_rcvd  = 0;
        esp_entry->ntimeouts = 0;
    } else {
        esp_entry->ntimeouts++;
        if (esp_entry->ntimeouts >= ESP_TMOUT_COUNT) {
            esp_free_entry(data);
            return;
        }
    }
    esp_entry->refresh_timer.expires = jiffies + ESP_REF_TMOUT;
    esp_entry->refresh_timer.function = esp_refresh_ct;
    esp_entry->refresh_timer.data = data;
    add_timer(&esp_entry->refresh_timer);
    esp_entry->timer_active = 1;
    DEBUGP("Refreshed timer pkt_rcvd %d timeouts %d\n",
           esp_entry->pkt_rcvd, esp_entry->ntimeouts);
}

/*
 * Allocate a free IPSEC table entry.
 */
struct _esp_table *alloc_esp_entry ( void )
{
	int idx = 0;
	struct _esp_table *esp_entry = esp_table;

	for ( ; idx < MAX_PORTS; idx++ ) {
		if ( esp_entry->inuse == IPSEC_FREE ) {
			esp_entry->tspi  = cur_spi = TEMP_SPI_START + idx;
			esp_entry->inuse = IPSEC_INUSE;
			DEBUGP( "New esp_entry at idx %d entry %p tspi %u cspi %u\n",
					idx, esp_entry, esp_entry->tspi, cur_spi );
			init_timer(&esp_entry->refresh_timer);
			esp_entry->refresh_timer.data     = idx;
			esp_entry->pkt_rcvd               = 0;
			esp_entry->refresh_timer.expires  = jiffies + ESP_REF_TMOUT;
			esp_entry->refresh_timer.function = esp_refresh_ct;
			add_timer(&esp_entry->refresh_timer);
			esp_entry->timer_active         = 1;
			return esp_entry;
		}
		esp_entry++;
	}
	return NULL;
}

/*
 * Search an ESP table entry by the Security Parameter Identifier (SPI).
 */
struct _esp_table *search_esp_entry_by_spi ( const struct esphdr *esph,
					     u_int32_t saddr, u_int32_t daddr )
{
	int idx = 0;
	struct _esp_table *esp_entry = esp_table;

	DEBUGP( "(0x%x) %u.%u.%u.%u %u.%u.%u.%u\n", 
			ntohl(esph->spi), NIPQUAD(saddr), NIPQUAD(daddr) );

	for ( ; idx < MAX_PORTS; idx++, esp_entry++ ) {
		if ( esp_entry->inuse == IPSEC_FREE ) {
			continue;
		}
		/* If we have seen traffic both ways */
		if ( esp_entry->r_spi != 0 ) {
			if ( esp_entry->l_spi == ntohl(esph->spi) ||
				 esp_entry->r_spi == ntohl(esph->spi) ) {
				DEBUGP("Both Ways Traffic Entry %p\n", esp_entry);
				return esp_entry;
			}
			continue;
		}

		/* If we have seen traffic only one way */
		if ( esp_entry->r_ip == ntohl(saddr) || esp_entry->l_ip == ntohl(daddr) ) {
			/* This must be the first packet from remote */
			esp_entry->r_spi = ntohl(esph->spi);
			DEBUGP("First Packet from Remote Entry %p\n", esp_entry);
			return esp_entry;
		}
		if ( ntohl(esph->spi) == esp_entry->l_spi ) {
			DEBUGP("One Way Traffic From Local Entry %p\n", esp_entry);
			return esp_entry;
		}
	}
	DEBUGP("No Entry\n");
	return NULL;
}

static int esp_pkt_to_tuple(const struct sk_buff *skb, unsigned int dataoff,
			    struct ip_conntrack_tuple *tuple)
{
	struct esphdr esph;
	struct _esp_table *esp_entry;
	const struct iphdr *iph = skb->nh.iph;

	if (skb_copy_bits(skb, dataoff, &esph, sizeof(esph)) != 0)
		return 0;

	DEBUGP("(0x%x) IP Pkt Hdr %u.%u.%u.%u <-> %u.%u.%u.%u\n",
           ntohl(esph.spi), NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
	DEBUGP("(0x%x) %u.%u.%u.%u <-> %u.%u.%u.%u\n", 
           ntohl(esph.spi), NIPQUAD(tuple->src.ip), NIPQUAD(tuple->dst.ip) );

	esp_entry = search_esp_entry_by_spi(&esph, tuple->src.ip, tuple->dst.ip);
	if (esp_entry == NULL ) {
		esp_entry = alloc_esp_entry();
		if ( esp_entry == NULL ) {
			return 0;
		}
		esp_entry->l_spi = ntohl(esph.spi);
		esp_entry->l_ip  = ntohl(tuple->src.ip);
		esp_entry->r_ip  = ntohl(tuple->dst.ip);
		esp_entry->r_spi = 0;
	}
	DEBUGP("tspi %u cspi %u spi 0x%x seq 0x%x sip %u.%u.%u.%u dip %u.%u.%u.%u\n",
		   esp_entry->tspi, cur_spi, ntohl(esph.spi), ntohl(esph.seq),
		   NIPQUAD(tuple->src.ip), NIPQUAD(tuple->dst.ip) );
	tuple->dst.u.esp_spi = esp_entry->tspi;
	tuple->src.u.esp_spi = esp_entry->tspi;
	esp_entry->pkt_rcvd++;
	return 1;
}

static int esp_invert_tuple(struct ip_conntrack_tuple *tuple,
			    const struct ip_conntrack_tuple *orig)
{
	DEBUGP("cspi 0x%x dspi 0x%x sspi 0x%x %u.%u.%u.%u <-> %u.%u.%u.%u\n",
			cur_spi, orig->dst.u.esp_spi, orig->src.u.esp_spi,
			NIPQUAD(tuple->src.ip), NIPQUAD(tuple->dst.ip) );

	tuple->src.u.esp_spi = orig->dst.u.esp_spi;
	tuple->dst.u.esp_spi = orig->src.u.esp_spi;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static int esp_print_tuple(struct seq_file *s,
				    const struct ip_conntrack_tuple *tuple)
{
	return seq_printf(s, "sport=%u dport=%u ",
		       ntohs(tuple->src.u.esp_spi), ntohs(tuple->dst.u.esp_spi));
}

/* Print out the private part of the conntrack. */
static int esp_print_conntrack(struct seq_file *s,
					const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static int esp_packet(struct ip_conntrack *conntrack,
		      const struct sk_buff *skb,
		      enum ip_conntrack_info conntrackinfo)
{
	const struct iphdr *iph = skb->nh.iph;
	const struct esphdr *esph = (void *)iph + iph->ihl*4;

	DEBUGP("(0x%x) %u.%u.%u.%u <-> %u.%u.%u.%u %s\n",
			ntohl(esph->spi), NIPQUAD(iph->saddr), NIPQUAD(iph->daddr),
			(conntrackinfo == IP_CT_NEW ) ? "CT_NEW" : "SEEN_REPLY" );
	DEBUGP("conntrackinfo %d\n", conntrackinfo);
	DEBUGP("status %lu\n", conntrack->status);

	if (conntrack->status & IPS_SEEN_REPLY) {
		ip_ct_refresh(conntrack, skb, ESP_CONN_TMOUT);
		set_bit(IPS_ASSURED_BIT, &conntrack->status);
	} else {
		ip_ct_refresh(conntrack, skb, ESP_REF_TMOUT);
   	}

	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int esp_new(struct ip_conntrack *conntrack, const struct sk_buff *skb)
{
	const struct iphdr *iph = skb->nh.iph;
	const struct esphdr *esph = (void *)iph + iph->ihl*4;

	DEBUGP("(0x%x) %u.%u.%u.%u <-> %u.%u.%u.%u\n",
			ntohl(esph->spi), NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
	return 1;
}

struct ip_conntrack_protocol ip_conntrack_protocol_esp = {
	.proto           = IPPROTO_ESP,
	.name            = "esp",
	.pkt_to_tuple    = esp_pkt_to_tuple,
	.invert_tuple    = esp_invert_tuple,
	.print_tuple     = esp_print_tuple,
	.print_conntrack = esp_print_conntrack,
	.packet          = esp_packet,
	.new             = esp_new,
	.me              = THIS_MODULE,
};

static int __init init(void)
{
    if (ip_conntrack_protocol_register(&ip_conntrack_protocol_esp)) {
        return -EIO;
    }
	printk("ip_conntrack_proto_esp loaded\n");

    return 0;
}

static void __exit fini(void)
{
    ip_conntrack_protocol_unregister(&ip_conntrack_protocol_esp);
}

module_init(init);
module_exit(fini);

