/***********************************************************************
*
* The Rome was not built in a day, keep moving 2007! :-)
*
* PPPD plugin for kernel-mode PPTP on Linux
*
* Copyright (C) 2007 by Delta Network Inc., haiyue wang
*
* Borrows heavily from the PPPoE plugin by Michal Ostrowski and Jamal Hadi Salim.
* and PPTP-Linux-1.1.0 by C. Scott Ananian.
*
***********************************************************************/
#include "pptp.h"

#include "pppd/pppd.h"
#include "pppd/fsm.h"
#include "pppd/lcp.h"
#include "pppd/ipcp.h"
#include "pppd/pathnames.h"

#include <linux/types.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <net/if.h>
#include <netdb.h>

extern int new_style_driver;
extern int session_ctl_fd;
extern int unsuccess;
extern int retry_time;

char pppd_version[] = VERSION;

static char *pptp_dns = NULL; /* Used to resolve the FQDN of the PPTP server */
static char *pptp_gateway = NULL; /* Used to get My IP Address and connect to PPTP server if not in the same subnet */
static char *pptp_iface = NULL;	/* Which interface are we Running */

/*
 * [NETGEAT SPEC V1.6] 5.3 PPTP
 *   The PPTP Server Address can be either an IP address or a FQDN. If the Server Address is a FQDN, the My IP
 * Address MUST be blank to indicate DHCP is used to get the router's IP address for establishing the PPTP tunnel.
 * So the router can acquire the DNS server address(es) to resolve the FQDN of the PPTP server. Besides, after the
 * PPTP server FQDN has been resolved, the router MUST add a static host route to the PPTP server and the gateway
 * of the route is Gateway IP Address user specified manually or the IP address specified in the Router option of
 * the DHCP Offer if DHCP is used to get My IP Address. If there are more than one IP address mapping to the PPTP
 * server FQDN, only the IP address the router uses to establish the PPTP link needs the static route. This static
 * route MUST be flushed when the PPTP link is down.
 *   In the case when the router acts as a DHCP client to get the IP address for establishing the PPTP link, following
 * requirements MUST be met.
 *   The DNS server address(es) acquired through DHCP MUST be and only be used by the router to query the IP address
 * of the PPTP server domain name, if a server FQDN is used. For all other queries, including the queries relayed for
 * a LAN PC, the DNS server(s) assigned from the PPTP connection MUST be used.
 */
struct rtentry *rt_pptpserver;	/* The static host route to PPTP server if needed */
struct rtentry *rt_pptdnssvr;	/* The static host route to DNS nameserver if needed */	

static int pptp_plugin_hook(char *cmd, char **argv, int doit);

static option_t pptp_options[] = {
	 /* we use 'devnam' to save our PPTP server address. */
	{ "Server Address", o_wild, (void *) &pptp_plugin_hook, "PPTP Server Address",
	   OPT_INITONLY | OPT_PRIVFIX | OPT_NOARG  | OPT_A2STRVAL | OPT_STATIC, devnam },
	{ "pptp_dns", o_string, &pptp_dns, "Used to resolve the FQDN of the PPTP server" },
	{ "pptp_gateway", o_string, &pptp_gateway, "Used to get My IP Address and connect to PPTP server" },
	{ "pptp_iface", o_string, &pptp_iface, "PPTP running on this interface" },
	{ NULL }
};

static struct pptp_conn_mngr *pptp_conn;

#define NIPQUAD(addr)	\
	((unsigned char *)&addr)[0],	\
	((unsigned char *)&addr)[1],	\
	((unsigned char *)&addr)[2],	\
	((unsigned char *)&addr)[3]

/* The bellowing looks ugly, but works well with 'NETGEAR SPEC V1.6', KeeP MovinG. -- 20090111 :) */

#define PPTP_DNS_TMPFILE	"/tmp/pptp_resolv.conf"

#define sin_addr(s) (((struct sockaddr_in *)(s))->sin_addr.s_addr)

static struct rtentry *setup_static_route(unsigned int destip)
{
	int fd;
	struct ifreq ifr;
	struct rtentry *rt;
	unsigned int netmask, ipaddr;

	if (pptp_gateway == NULL || pptp_iface == NULL) {
		/* printf("PPTP Gateway or PPTP running interface is NUL!\n"); */
		return NULL;
	}

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
		return NULL;

	memset(&ifr, 0, sizeof(struct ifreq));
	ifr.ifr_addr.sa_family = AF_INET;
	strcpy(ifr.ifr_name, pptp_iface);

	if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) 
		ipaddr = sin_addr(&ifr.ifr_addr);
	else
		ipaddr = 0;

	if (ioctl(fd, SIOCGIFNETMASK, &ifr) == 0)
		netmask = sin_addr(&ifr.ifr_netmask);
	else
		netmask = 0;

	close(fd);

	printf("Destination IP: %u.%u.%u.%u ---> PPTP Ethernet Interface: %s %u.%u.%u.%u/%u.%u.%u.%u\n",
		NIPQUAD(destip), pptp_iface, NIPQUAD(ipaddr), NIPQUAD(netmask));

	if ((destip & netmask) == (ipaddr & netmask)) {
		printf("The same subnet, need no static host route!!!\n");
		return NULL;
	}

	rt = malloc(sizeof(struct rtentry));
	if (rt != NULL) {
		printf("NOT in the same subnet, need a static host route...\n");
		memset(rt, 0, sizeof(struct rtentry));

		rt->rt_metric = 1;
		rt->rt_dev = pptp_iface;
		rt->rt_flags = RTF_UP | RTF_HOST | RTF_GATEWAY;
		sin_addr(&rt->rt_dst) = destip;
		sin_addr(&rt->rt_genmask) = 0xFFFFFFFF;
		sin_addr(&rt->rt_gateway) = inet_addr(pptp_gateway);
		rt->rt_dst.sa_family = AF_INET;
		rt->rt_gateway.sa_family = AF_INET;
		rt->rt_genmask.sa_family = AF_INET;
	}

	return rt;
}

/* Add or delete a route */
static void maintain_route(struct rtentry *rt, int add)
{
	int sock;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
		return;

	ioctl(sock, add ? SIOCADDRT : SIOCDELRT, rt);
	close(sock);
}

static void setup_pptp_nameserver(void)
{
	FILE *fp;

	rt_pptdnssvr = setup_static_route(inet_addr(pptp_dns));
	if (rt_pptdnssvr != NULL)
		maintain_route(rt_pptdnssvr, 1);

	if (!usepeerdns)
		rename(_PATH_RESOLV, PPTP_DNS_TMPFILE);
	fp = fopen(_PATH_RESOLV, "w");
	if (fp == NULL)
		return;
	fprintf(fp, "nameserver	%s\n", pptp_dns);
	fclose(fp);
}

static void flush_pptp_nameserver(void)
{
	if (!usepeerdns) {
		unlink(_PATH_RESOLV);
		rename(PPTP_DNS_TMPFILE, _PATH_RESOLV);
	}

	if (rt_pptdnssvr != NULL) {
		maintain_route(rt_pptdnssvr, 0);
		free(rt_pptdnssvr);
		rt_pptdnssvr = NULL;
	}
}

static unsigned int get_pptp_server(void) 
{
	unsigned int addr;
	struct hostent *host;

	addr = inet_addr(devnam);
	/* It is an IP address. */
	if (addr != INADDR_NONE) {
		rt_pptpserver = setup_static_route(addr);
		return addr;
	}

	/* It is a FQDN, needs DNS Server to resolve it. */
	addr = 0;
	if (pptp_dns == NULL)
		pptp_fatal("The PPTP Server is FQDN, but no DNS nameserver configured!");

	setup_pptp_nameserver();
	host = gethostbyname(devnam);
	if (host != NULL && host->h_addrtype == AF_INET)
		memcpy(&addr, host->h_addr, sizeof(addr));
	flush_pptp_nameserver();

	if (addr != 0) {
		printf("Resolved PPTP Server \"%s\" to [%u.%u.%u.%u] successfully!\n", devnam, NIPQUAD(addr));
		rt_pptpserver = setup_static_route(addr);
	}

	return addr;
}

static void init_pptp(void)
{
	struct pptp_conn_mngr * conn;
	
	conn = malloc(sizeof(struct pptp_conn_mngr));

	if (conn == NULL)
		pptp_fatal("Can't allocate memory for PPTP session");

	memset(conn, 0, sizeof(struct pptp_conn_mngr));

	conn->call_socket = -1;
	conn->gre_socket = -1;

	conn->conn_state = CONN_IDLE;
	conn->call_state = CALL_IDLE;

	//conn->call_id = (u16)time(NULL);
	conn->call_sernum = 0x2312;

	pptp_conn = conn;
}

/* If the PPTP server is NOT in the same subnet with DUT, it needs a static host route. */
static void setup_conn_route(void)
{
	if (rt_pptpserver != NULL)
		maintain_route(rt_pptpserver, 1);
}

static void flush_conn_route(void)
{
	if (rt_pptpserver != NULL)
		maintain_route(rt_pptpserver, 0);
}

static int pptp_open_session(void)
{
	session_ctl_fd = -1;
	struct pptp_conn_mngr *conn = pptp_conn;
	struct sockaddr_pptp sp;

	//conn->call_id++;
	conn->call_sernum++;

	strcpy(ppp_devnam, "gre0");
#if 0
	if (pptp_dns != NULL)
		pptp_info("PPTP DNS: %s", pptp_dns);
	if (pptp_gateway != NULL)
		pptp_info("PPTP Gateway: %s", pptp_gateway);
	if (pptp_iface != NULL)
		pptp_info("PPTP Interface: %s", pptp_iface);
#endif
	if ((conn->svr_addr.s_addr == 0) && (conn->svr_addr.s_addr = get_pptp_server()) == 0) {
		fprintf(stderr, "Can't resolve the PPTP Server address: %s\n", devnam);
		return -1;
	}

	setup_conn_route();
	open_callmngr(conn);

	if (conn->call_state != CALL_ESTABLISHED) {
		fprintf(stderr, "Unable to complete PPTP Control-Call Connection!\n");
		flush_conn_route();
		return -1;
	}

	pptp_info("PPTP Call-ID: %u, Peer-Call-ID: %u", conn->call_id, 
			conn->peer_call_id);
	
	conn->gre_socket = socket(AF_INET, SOCK_GRE_PPP, IPPROTO_GRE);
	if (conn->gre_socket < 0)
		pptp_fatal("Failed to create PPTP socket: %m");
		
	sp.sa_family = AF_INET;
	sp.pa.call_id = htons(conn->call_id);
	sp.pa.peer_callid = htons(conn->peer_call_id);
	sp.pa.peer_addr = conn->svr_addr.s_addr;
	
	if (connect(conn->gre_socket, (struct sockaddr *) &sp, sizeof(struct sockaddr_pptp)) < 0)
		pptp_fatal("Failed to connect PPTP socket: %d %m", errno);

	session_ctl_fd = conn->call_socket;

	return conn->gre_socket;
}


static void pptp_send_config(int mtu, u32 asyncmap, int pcomp, int accomp)
{
	if (mtu > MAX_PPTP_MTU)
		pptp_info("Couldn't increase MTU to %d", mtu);
}


static void pptp_recv_config(int mru, u32 asyncmap, int pcomp, int accomp)
{
	if (mru > MAX_PPTP_MTU)
		pptp_info("Couldn't increase MRU to %d", mru);
}

static void pptp_close_session(void)
{
	struct pptp_conn_mngr * conn = pptp_conn;

	/* Maybe it has been disconnected by server. */
	if (conn->call_state == CALL_ESTABLISHED)
		close_callmngr(conn);
	
	close(conn->gre_socket);
	close(conn->call_socket);

	conn->gre_socket = -1;
	conn->call_socket = -1;

	conn->call_state = CALL_IDLE;
	conn->conn_state = CONN_IDLE;

	flush_conn_route();
}

static void pptp_input_packet(void)
{
	int ret = -1;
	struct pptp_conn_mngr * conn = pptp_conn;

	
	if (conn->call_state == CALL_ESTABLISHED)
		ret = pptp_get_input(conn);

	if (ret == PPTP_RET_TERM_LINK) {
		link_terminated(0);

		/* Let's set the bellowing value to connect server soon!*/
		unsuccess = 1;
		retry_time = 1;
	}
}

void pptp_time_out(void * arg)
{
	struct pptp_conn_mngr * conn = (struct pptp_conn_mngr *)arg;

	if (conn->call_state == CALL_ESTABLISHED) {
		pptp_handle_timer(conn);

		/* Repeat!*/
		TIMEOUT(pptp_time_out, conn, 59);
	}
}

void pptp_timer(int on)
{
	struct pptp_conn_mngr * conn = pptp_conn;

	if (on && conn->call_state == CALL_ESTABLISHED)
		TIMEOUT(pptp_time_out, conn, 30);

	/* Just try to do it!*/
	if (on == 0)
		UNTIMEOUT(pptp_time_out, conn);
}

struct channel pptp_channel = {
	.options	= pptp_options,
	.process_extra_options	= NULL,
	.check_options			= NULL,
	.connect				= &pptp_open_session,
	.disconnect			= &pptp_close_session,
	.establish_ppp			= &generic_establish_ppp,
	.disestablish_ppp		= &generic_disestablish_ppp,
	.send_config			= &pptp_send_config,
	.recv_config			= &pptp_recv_config,
	.close				= NULL,
	.cleanup				= NULL,
	.input_packet			= &pptp_input_packet,
	.handle_timer			= &pptp_timer,
};

static int pptp_plugin_hook(char *cmd, char **argv, int doit)
{
	strncpy(devnam, cmd, sizeof(devnam));
	
	if (the_channel != &pptp_channel) {
		the_channel = &pptp_channel;

		modem = 0;

		lcp_allowoptions[0].neg_accompression = 0;
		lcp_wantoptions[0].neg_accompression = 0;

		lcp_allowoptions[0].neg_asyncmap = 0;
		lcp_wantoptions[0].neg_asyncmap = 0;

		lcp_allowoptions[0].neg_pcompression = 0;
		lcp_wantoptions[0].neg_pcompression = 0;

		ipcp_allowoptions[0].neg_vj=0;
		ipcp_wantoptions[0].neg_vj=0;

		init_pptp();
	}

	return 1;
}


void plugin_init(void)
{
	if (!ppp_available() && !new_style_driver)
		pptp_fatal("Linux kernel does not support PPTP -- are you running 2.4.x?");

	add_options(pptp_options);
}

